#!/usr/bin/perl -w
# OverCR Collector V1.50.00 08/15/1999
#
# This is GPL software, see COPYING for liscense
# Copyright 1998, 1999
#
#  People who have contributed code
#
#  Eric Molitor  (eric@molitor.org)
#  Bill Jones    (bill@fccj.org) AKA -Sneex-
#  Joerg Strangfeld <joerg.strangfeld@javelinx.com>
#

# 1.49.08 fixed so far
#
#   1.49.07 - Fixed Glob Failure caused by forking
#   1.49.06 - Better Handling of Config File Location
#   1.49.05 - Daemonize with code from Joerg Strangfeld
#   1.49.04 - Signal HUP is now processed to reconfigure
#   1.49.03 - Spelling and logging format fixes
#   1.49.02 - Implemented allowed services in config file
#   1.49.02 - Minor Commenting and formatting changes
#   1.49.01 - Require 5.004
#   1.49.01 - Added Config File Support
#   1.00.00 - Initial Release
#   0.99.53 - Rewrote Uptime
#   0.99.52 - Improved Uptime Help Information
#   0.99.52 - Added postel function
#   0.99.51 - Added help function
#   0.99.51 - Added architecture to version
#   0.99.51 - Fixed netstat on Solaris/SunOS
#   0.99.50 - Removed/Changed reserved words
#   0.99.50 - Removed D.O.S. attack (Empty Commands now handled correctly)
#   0.99.50 - Fixed signal handling of SigAlarm
#   0.99.50 - Check return values on ALL IO operations
#   0.99.50 - Fixed orphaned die during socket creation
#   0.99.50 - Cleaned up variables in ocr_uptime
#   0.99.50 - Uptime now has a precision of 2
#   0.99.50 - Fixed Netstats and Process whining about uninitialized variables

# To Do
# - Improved architecture detection and handling

# <CURRENT CODE BASE: 1.50.00>
# End of comments...

require 5.004;

# Uncomment these two during development only...
use strict;
use diagnostics;

use Socket;
use Carp;
use Config;
use POSIX;
use IO;


# Dont change anything below!
# (Unless you want to become an OverCR Developer!)


# Global Declerations
#
#

# Shall we report Debug Messages?
my $ocr_DEBUG=1;

# Configuration Variables
my @supported_commands =
   ('uptime','load','diskspace','postel','process','netstat','help');

my $ConfigFile;
my @allowed_commands;
my @ip_list;
my $LogFile;
my $port;
my $SOCKET_TIMEOUT;

my $version="1.49.08";
my $quit=0;
my $sigstable;			# Do we need to reinstall signal handlers

sub spawn;			# Forward decleration for forking


# Detect Operating System
# ugly, but effective.
#
my $opsys = lc($^O) || lc(( split (/ /, `uname -a` ))[0]);
$opsys = "sunos" if ($^O =~ /solaris/); # Cover both sunos and solaris

sub process_config
# Configuration File
# Rev. 1
#
{
   my @ocr_line_array;

   @allowed_commands='';
 
   if (open OCR_CONFIG, "/etc/collector.conf") {
      $ConfigFile="/etc/collector.conf";
   } elsif (open OCR_CONFIG, "collector.conf") {
      $ConfigFile="./collector.conf";
   } else {
      die "Config file not found!";
   }

   my $ocr_line='';
 
   # Process comments and ';' delimited lines.
   while (<OCR_CONFIG>) {
      my $temp_line=$_;
      chomp($temp_line);
      if ($temp_line=~/#/) {
         $temp_line=~s/#.*$//;
      }
      if ($temp_line=~/;/) {
         $ocr_line= join '', $ocr_line, (split (/;/,$temp_line))[0];
         push (@ocr_line_array, $ocr_line);
         $ocr_line='';
         $temp_line=~s/^.*;//;
      } else {
         $ocr_line= join '', $ocr_line, $temp_line;
      }
   }
   close OCR_CONFIG;
 
   foreach my $temp_array (@ocr_line_array) {
      if ($temp_array=~/services/) {
         $temp_array=~s/services//;
         $temp_array=~s/(\,|{|})/ /g;
         # remove leading spaces
         $temp_array=~s/^ +//;
         my @temp_command=split (/ /,$temp_array);
         while (<@temp_command>) {
            my $found=0;
            foreach my $temp_service (@supported_commands) {
               if ($_ eq $temp_service) {
                  $found=1;
                  push (@allowed_commands, $temp_service);
               }
            }                   
            if (!$found) {
               print "Unknown Service In Services:\n$_\n";
               exit;
            }
            
         }
      } elsif ($temp_array=~/ip_allow/) {
         $temp_array=~s/ip_allow//;
         $temp_array=~s/(\,|{|})/ /g;
         # remove leading spaces
         $temp_array=~s/^ +//;
         my @temp_ip_array = split (/ /,$temp_array);
         while (<@temp_ip_array>) {
            # Verify that it is a valid IP address ... whew!
            if (/(?!0+\.0+\.0+\.0+$)^([01]?\d\d?|2[0-4]\d|25[0-5])\.([01]?\d\d?|2[0-4]\d|25[0-5])\.([01]?\d\d?|2[0-4]\d|25[0-5])\.([01]?\d\d?|2[0-4]\d|25[0-5])$/) {
               push (@ip_list, $_);
            } else {
               print "Invalid IP address in ip_allow portion of config file!:\n$_\n";
               exit;
            }
         }
      } elsif ($temp_array=~/port/) {
         if ($temp_array=~/=/) {
            my $temp_command = (split (/=/,$temp_array))[1];
            $temp_command=~s/^[\s]*//;
            $temp_command=~s/[\s]*$//;
            if ($temp_command=~/[\D]/) {
               print "port must be a number!:\n[$temp_command]\n";
               exit;
            } else {
               $port = shift || $temp_command;
            }
         } else {
            print "Improper Command Format!:\n $temp_array\n";
         }
      } elsif ($temp_array=~/timeout/) {
         if ($temp_array=~/=/) {
            my $temp_command = (split (/=/,$temp_array))[1];
            $temp_command=~s/^[\s]*//;
            $temp_command=~s/[\s]*$//;
            if ($temp_command=~/[\D]/) {
               print "Timeout must be a number!:\n[$temp_command]\n";
               exit;
            } else {
               $SOCKET_TIMEOUT=$temp_command;
            }
         } else {
            die "Improper Command Format!:\n $temp_array\n";
         }
      } elsif ($temp_array=~/log_file/) {
         if ($temp_array=~/=/) {
            my $temp_command = (split (/=/,$temp_array))[1];
            $temp_command=~s/("|')//g;
            $temp_command=~s/^[\s]*//;
            $LogFile=$temp_command;
         } else {
            print "Improper Command Format!:\n $temp_array\n";
         }
      } else {
         print "Unknown Command in Line!:\n$temp_array\n";
         exit;
      }
   }
}
 

sub logmsg
# Log Management routine
# (rewritten by -Sneex- January 31st, 1999 @ 10:00am...)
#
{
   # Open it Write Only (O_WRONLY), in Append Mode (O_APPEND),
   # if it already exists; otherwise Create the file (O_CREAT)
   # Set the file permission bits to 0600  ( -rw------- )
   my $fh = new IO::File $LogFile, O_WRONLY|O_APPEND|O_CREAT, 0600;
 
   # Now simply output the notice info...
   if (defined $fh) {
      print $fh "$0 $$: @_ at ", scalar localtime, "\n";
      undef $fh;
   } else { print "Oops! Betcha that hurt!\nWhy: $!"; }
 
   # Flush all I/O buffers...
   autoflush STDOUT 1;
}


sub allow_service
# Determines if a service is allowed to be used.
#
{
   my $test=0;
   foreach my $temp_allow (@allowed_commands) {
      if ($temp_allow eq ($_[0])) {
         $test=1;
      }
   }
   return $test;
}




sub ocr_postel
# Tribute to Dr. Jonathan B. Postel
#
#
{
   if (allow_service('postel')) {
      print "Dr. Jonathatn B. Postel, RIP October 16, 1998\n";
   } else {
      print "Disabled\n";
   }
   print "\n";
}



sub ocr_load
# Reports System Load
#
#
{
   if (allow_service('load')) {
      open UPTIME, "/usr/bin/uptime|" or die "Oops! $!"; # Open uptime
      my $load=<UPTIME>; # Dump it all into a variable
      close UPTIME;
      $load=(split(/average: /,$load))[1];
      chomp $load;
      my @load=split(/, /,$load);
      foreach my $temp_load (@load) {
         print "$temp_load\n";
      }
   } else {
      print "Disabled\n";
   }
   print "\n";

}



sub ocr_uptime
# Reports System Uptime
#
#
{
   if (allow_service('uptime')) {
      open UPTIME, "/usr/bin/uptime|" or die "Oops! $!";
      my $uptime=<UPTIME>;
      close UPTIME;

      $uptime=~s/\d+ user.+$//;
      $uptime=~s/^.+up//;
      my @uptime=split (/, /, $uptime);
      $uptime=0;

      foreach my $temp (@uptime) {
         if ($temp=~/day/) {
            $temp=(split(/day/, $temp))[0];
            $uptime+=($temp*24)+$uptime;
         } elsif ($temp=~/:/) {
            $uptime+=(split(/:/,$temp))[0];
            $uptime+=((split(/:/,$temp))[1])/60;
         } elsif ($temp=~/min/) {
            $uptime+=((split(/min/,$temp))[0])/60;
         }
      }
      printf "%.2f\n", $uptime; # Limit to two decimal places
   } else {
      print "Disabled\n";
   }
   print "\n";
}



sub ocr_process
# Checks to see if processes exist on local machine
#
#
{
   if (allow_service('process')) {
      my @process_list;
      my @array;
      my $PSCommand;

      # Determine how to call PS
      if ($opsys eq "linux") {
         $PSCommand = "/bin/ps ax";
         # For Linux Systems
      } elsif ($opsys eq "sunos") {
         $PSCommand = "/bin/ps -ef";
         # For Sun Systems
      } else {
         $PSCommand = "/bin/ps -ef";
         # If other, then assume POSIX
      }

      my @check_list=split(/,/,$_[0]);

      open PS, "$PSCommand|" or die "Oops! $!"; # Open PS for input
      @process_list=<PS>; # Dump it all into an array
      close PS;

      foreach my $check (@check_list) {
         # Check everything in the list
         my $count=0; # Prime $count
         foreach my $process (@process_list) {
            if ($process=~/$check/) {
               $count+=1; # Hey, we found it!
            }
         }
         push (@array,"$check ($count)\n"); # Push the output array
      }

      print @array;  # Report Output array
   } else {
      print "Disabled\n";
   }
   print "\n";    # Give blank line to end section
}



sub ocr_diskspace
# Check diskspace on local machine
#
#
{
   if (allow_service('diskspace')) {
      my $DFCommand;
      my @array;

      # Determine how to call DF
      if ($opsys eq "linux") {
         $DFCommand = "/bin/df -P -k";
         # For Linux Systems
      } elsif ($opsys eq "sunos") {
         $DFCommand = "/bin/df -k";
         # For Sun Systems
      } else {
         $DFCommand = "/bin/df -k";
         # If other, then assume POSIX (Is this POSIX?)
      }

      open DFCOMMAND, "$DFCommand|" or die "Oops! $!";
      my $ignore=1; # Ugly, but it works for now
      while (<DFCOMMAND>) {
         if ($ignore) {
            $ignore=0; # Ignore first line
         }
         else {
            chomp; # Remove the newline
            (my $fsys,my $blocks,my $used,my $avail,my $cap,my $mount) =
               split(/ +/);
            push (@array,"$fsys $cap\n"); # Push onto the output array
            $cap=~s/%//; # Strip the percent sign
         }
      }
      close DFCOMMAND;

      print @array;  # Report output array
   } else {
      print "Disabled\n";
   }
   print "\n";    # Newline to end section
}



sub ocr_netstat
# Use Netstat to track connections
# Thanks to Nick Powers for this idea (nickp@unicomp.net)
#
{
   if (allow_service('netstat')) {
      my $count=0;
      my $port=$_[0];
      my $ignore=0;

      # OS Specific
      if ($opsys eq "linux") {
         # For Linux Systems
         open NETSTAT,"/bin/netstat -n --tcp|" or die "Oops! $!";
         while (<NETSTAT>) {
            # ignore the first 2 lines
            if ($ignore < 2) {
               $ignore+=1;
            } else {
               if ((split(/:/,(split(/ +/))[3]))[1] eq $port) {
                  $count+=1;
               }
            }
         }
         close NETSTAT;
      } elsif ($opsys eq "sunos") {
         # For Sun Systems
         open NETSTAT,"/bin/netstat -n -P tcp -f inet|" or die "Oops! $!";
         while (<NETSTAT>) {
            # ignore the first 4 lines
            if ($ignore < 4) {
               $ignore+=1;
            } else {
               if ((split(/\./,(split(/ +/))[0]))[4] eq $port) {
                  $count+=1;
               }
            }
         }
         close NETSTAT;
      } else {
         # If other, then do nothing for now.
      }
      print "$count\n";
   } else {
      print "Disabled\n";
   }
   print "\n";
}



sub ocr_help {
# Simple help routine
#
#
if (allow_service('help')) {
   print "Enter one of:

diskspace
process httpd,named,kerneld
netstat 2000
version
quit
help

COMMANDS
========

NETSTAT <port>
      Allows you to monitor the number of connections to a TCP/IP Port.
   EG: netstat 110
      Tells you the number of pop mail sockets.

PROCESS <process_name,process_name...>
      Allows you to monitor which processes are running.
   EG: process httpd,named
      Tells you if httpd and named are running and how many copies.

DISKSPACE
      Allows you to monitor file system utilization
   EG: diskspace
      Will tell you file system utilization for all local file systems.

UPTIME
      Allows you to monitor the systems uptime
   EG: uptime
      Will tell you the systems uptime.

LOAD
      Allows you to monitor system load
   EG: load
      Reports the load during the last 1, 5 and 15 minutes.

VERSION
      Notifies you of what version the OverCR collector is.
   EG: version
      Will tell you what version of the OverCR collector you are running.

POSTEL
      A tribute to the founder of the Internet.
   EG: postel
      Reports the date of death.

QUIT
      Quits, disconnects session.
   EG: quit

HELP
      Display this message.
   EG:  help\n";
   } else {
      print "Disabled\n";
   }
   print "\n";
}

sub sighup {
   logmsg "Recieved HUP Signal, reconfiguring";
   process_config();
}

##########################################################
# Begin Main Program
#
#

# Daemonize
# From Joerg Strangfeld (joerg.strangfeld@javelinx.com)

#
# Need a standard place to store the PID so we can stop
#    and start it easily.

my $daemon_pid = fork;
exit if $daemon_pid;
die "Couldn't fork: $!" unless defined($daemon_pid);
POSIX::setsid() || die "Can't start new session: $!";

$SIG{HUP}=\&sighup; # Install KILL signal handler

# Process Configuration File
process_config();

# Detect if we have stable signal handlers
$sigstable=1;
if ($Config{d_sigaction}) {
   $sigstable=0;
}

# Lets initialize the socket
# 
#
my $proto = getprotobyname('tcp'); # Use TCP
socket (Server, PF_INET, SOCK_STREAM, $proto) or die "socket: $!"; # Initialize Socket
setsockopt (Server, SOL_SOCKET, SO_REUSEADDR, pack ("l", 1))
    or die "setsockopt: $!"; # Set socket options
bind (Server, sockaddr_in($port, INADDR_ANY)) or die "bind: $!"; # Bind the Socket
listen (Server, SOMAXCONN) or die "listen: $!"; # Start it all up
logmsg "OverCR Collector Version $version";
logmsg "Configuration loaded from $ConfigFile";
logmsg "Server started on port $port";

# From this point on is where it gets ugly.
# You really dont need to know anything about the stuff down here
# other than it captures STDIN and STDOUT and sends it all through
# the socket.

my $waitedpid=0;
my $paddr;

sub REAPER {
# We shall reap what we sow (or at least fork)
   $waitedpid=wait;
   if (!$sigstable) {
      $SIG{CHLD} = \&REAPER; # Reinstall Signal Handler
   }
   if ($ocr_DEBUG) {
   logmsg "Reaped $waitedpid" . ($? ? " with exit $?" : "");
   };
}

$SIG{CHLD} = \&REAPER; # Install the signal handler

for ( $waitedpid=0;
      ($paddr = accept(Client, Server)) || $waitedpid;
      $waitedpid=0, close Client)
{
   next if $waitedpid and not $paddr;
   my ($port, $iaddr) = sockaddr_in($paddr);
   my $name = gethostbyaddr($iaddr,AF_INET);

    # Make sure we only talk to our trusted IP's
    #
   my $allow=1;
   foreach my $temp_ip (@ip_list) {
      if (inet_ntoa($iaddr) eq $temp_ip) {
         $allow=0;
      }
   }
   if ($allow ne 0) {
      # Lets log the offender
      logmsg "Connection attempt from untrusted host $name [",
      inet_ntoa($iaddr), "] at port $port";
   } else {

      if ($ocr_DEBUG) {
         # Log whomever is connecting to us
         logmsg "Connection from $name [", inet_ntoa($iaddr), "] on port $port";
      }

      spawn sub {
         # Look MA, we're bidirectional!
         #     - Eric
         # This is the main loop which captures the commands and calls the
         # appropriate subroutine.

         sub sigalarm {
            logmsg "Connection timeout during input from $name [",
               inet_ntoa($iaddr), "] on port $port";
            exit;
            # No need to reinstall the signal handler as we commit
            # suicide via die if we timeout.
         }

         $SIG{ALRM}=\&sigalarm; # Install the Signal Handler

         my $command_error=0;
         while (!$quit) {
            alarm $SOCKET_TIMEOUT; # Die if their are no commands
            my $command=<STDIN>;
            $command=~s/\n//; # Get rid of those pesky <CR>
            $command=~s/\r//; # Get rid of those pesky <LF>
            # Catch blank lines!
            if ($command ne '') {
               (my $action, my $argument) = split (/ /, $command);
               $action=~tr/a-z/A-Z/;
               if ($action eq "DISKSPACE") {
                  ocr_diskspace;
               } elsif ($action eq "UPTIME") {
                  ocr_uptime;
               } elsif ($action eq "LOAD") {
                  ocr_load;
               } elsif ($action eq "PROCESS") {
                  if ($argument) {
                     ocr_process($argument);
                  } else {
                     print "Missing required Argument(s)!\n\n";
                     $command_error+=1;
                  }
               } elsif ($action eq "NETSTAT") {
                  if ($argument) {
                     ocr_netstat($argument);
                  } else {
                     print "Missing required Argument(s)!\n\n";
                     $command_error+=1;
                  }
               } elsif ($action eq "POSTEL") {
                  ocr_postel;
               } elsif ($action eq "VERSION") {
                  print "$version - $opsys\n\n";
               } elsif ($action eq "HELP") {
                  &ocr_help;
               } elsif ($action eq "QUIT") {
                  $quit=1;
               } else {
                  print "ERROR: Unknown Command: {$command}\n";
                  $command_error+=1;
                  }
               } else {
                  # If it is a blank line, count it as an error.
                  # this removes a possible D.O.S. attack.
                  print "ERROR: Blank Command!\n";
                  $command_error+=1;
            }
            if ($command_error ge 5) {
               # Quit after 5 Errors
               $quit=1;
            }

         }
      }; # End spawn sub
   }
}



sub spawn 
# Here there be dragons!
#
#
{
   my $coderef = shift;

   unless (@_ == 0 && $coderef && ref($coderef) eq 'CODE') {
      confess "Usage: spawn CODEREF"; # Were we called correctly?
   }

   my $pid;
   if (!defined($pid=fork)) {
      logmsg "Cannot Fork: $!";
      return;
   } elsif ($pid) {
      if ($ocr_DEBUG) {
         logmsg "Begat $pid";
      }
      return;
   }

   # else I am the child, lets go spawn
   open(STDIN, "<&Client") or die "can't dup client to stdin"; # Capture Standard Input
   open(STDOUT, ">&Client") or die "can't dup client to stdout"; # Capture Standard Output
   STDOUT->autoflush();
   exit &$coderef();
}
